(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Abstract_Rewriting2
  imports Abstract_Rewriting
begin

lemma (in infinite_hits) index_not_p_start: assumes i: "i < index 0" shows "\<not> p i"
proof -
  from i[simplified index.simps] have "i < Least p" .
  from not_less_Least[OF this] show ?thesis .
qed

(* TODO: this lemma is more general than Newman and should
  replace Newman in Abstract_Rewriting *)
lemma Newman_local:
  assumes "SN_on r X" and WCR: "WCR_on r {x. SN_on r {x}}"
  shows "CR_on r X"
proof - {
  fix x
  assume "x \<in> X"
  with assms have "SN_on r {x}" unfolding SN_on_def by auto
  with this  have "CR_on r {x}"
  proof (induct rule: SN_on_induct')
    case (IH x) show ?case
    proof
      fix y z assume "(x,y) \<in> r^*" and "(x,z) \<in> r^*"
      from `(x,y) \<in> r^*` obtain m where "(x,y) \<in> r^^m" ..
      from `(x,z) \<in> r^*` obtain n where "(x,z) \<in> r^^n" ..
      show "(y,z) \<in> r\<^sup>\<down>"
      proof (cases n)
        case 0
        from `(x,z) \<in> r^^n` have eq: "x = z" by (simp add: 0)
        from `(x,y) \<in> r^*` show ?thesis unfolding eq ..
      next
        case (Suc n')
        from `(x,z) \<in> r^^n`[unfolded Suc] obtain t where "(x,t) \<in> r" and "(t,z) \<in> r^*" ..
        show ?thesis
        proof (cases m)
          case 0
          from `(x,y) \<in> r^^m` have eq: "x = y" by (simp add: 0)
          from `(x,z) \<in> r^*` show ?thesis unfolding eq ..
        next
          case (Suc m')
          from `(x,y) \<in> r^^m`[unfolded Suc] obtain s where "(x,s) \<in> r" and "(s,y) \<in> r^*" ..          
          from WCR IH(2) have "WCR_on r {x}" unfolding WCR_on_def by auto
          with `(x,s) \<in> r` and `(x,t) \<in> r` have "(s,t) \<in> r\<^sup>\<down>" by auto
          then obtain u where "(s,u) \<in> r^*" and "(t,u) \<in> r^*" ..
          from `(x,s) \<in> r` IH(2) have "SN_on r {s}" by (rule step_preserves_SN_on)
          from IH(1)[OF `(x,s) \<in> r` this] have "CR_on r {s}" .
          from this and `(s,u) \<in> r^*` and `(s,y) \<in> r^*` have "(u,y) \<in> r\<^sup>\<down>" by auto
          then obtain v where "(u,v) \<in> r^*" and "(y,v) \<in> r^*" ..
          from `(x,t) \<in> r` IH(2) have "SN_on r {t}" by (rule step_preserves_SN_on)
          from IH(1)[OF `(x,t) \<in> r` this] have "CR_on r {t}" .
          moreover from `(t,u) \<in> r^*` and `(u,v) \<in> r^*` have "(t,v) \<in> r^*" by auto
          ultimately have "(z,v) \<in> r\<^sup>\<down>" using `(t,z) \<in> r^*` by auto
          then obtain w where "(z,w) \<in> r^*" and "(v,w) \<in> r^*" ..
          from `(y,v) \<in> r^*` and `(v,w) \<in> r^*` have "(y,w) \<in> r^*" by auto
          with `(z,w) \<in> r^*` show ?thesis by auto
        qed
      qed
    qed
  qed
  }
  thus ?thesis unfolding CR_on_def by blast
qed

lemma SN_reaches_NF: assumes "SN_on r {x}"
  shows "\<exists> y. (x,y) \<in> r^* \<and> y \<in> NF r"
  using assms
proof (induct rule: SN_on_induct')
  case (IH x)
  show ?case
  proof (cases "x \<in> NF r")
    case True
    thus ?thesis by auto
  next
    case False
    then obtain y where step: "(x,y) \<in> r" by auto
    from IH[OF this] obtain z where steps: "(y,z) \<in> r^*" and NF: "z \<in> NF r" by auto
    show ?thesis
      by (intro exI, rule conjI[OF _ NF], insert step steps, auto)
  qed
qed

lemma SN_WCR_reaches_NF: assumes SN: "SN_on r {x}" 
  and WCR: "WCR_on r {x. SN_on r {x}}"
  shows "\<exists>! y. (x,y) \<in> r^* \<and> y \<in> NF r"
proof -
  from SN_reaches_NF[OF SN] obtain y where steps: "(x,y) \<in> r^*" and NF: "y \<in> NF r" by auto
  show ?thesis
  proof(rule, rule conjI[OF steps NF])
    fix z
    assume steps': "(x,z) \<in> r^* \<and> z \<in> NF r"
    from Newman_local[OF SN WCR] have "CR_on r {x}" by auto
    from CR_onD[OF this _ steps] steps' have "(y,z) \<in> r\<^sup>\<down>" by simp
    from join_NF_imp_eq[OF this NF] steps' show "z = y" by simp
  qed
qed

definition some_NF :: "'a ars \<Rightarrow> 'a \<Rightarrow> 'a"
  where "some_NF r x \<equiv> SOME y. (x,y) \<in> r^* \<and> y \<in> NF r"

lemma some_NF: assumes SN: "SN_on r {x}" 
  shows "(x,some_NF r x) \<in> r^* \<and> some_NF r x \<in> NF r"
  using someI_ex[OF SN_reaches_NF[OF SN]]
  unfolding some_NF_def .

lemma the_NF: assumes SN: "SN_on r {x}"
  and WCR: "WCR_on r {x. SN_on r {x}}"
  and steps: "(x,y) \<in> r^*"
  and NF: "y \<in> NF r"
  shows "y = some_NF r x"
proof -
  let ?p = "\<lambda> y. (x,y) \<in> r^* \<and> y \<in> NF r"
  from SN_WCR_reaches_NF[OF SN WCR]
  have one: "\<exists>! y. ?p y" .
  from steps NF have y: "?p y" ..
  from some_NF[OF SN] have some: "?p (some_NF r x)" .
  from one some y show ?thesis by auto
qed

lemma the_NF_UNF: assumes UNF: "UNF r"
  and steps: "(x,y) \<in> r^*"
  and NF: "y \<in> NF r"
  shows "y = some_NF r x"
proof -
  let ?p = "\<lambda> y. (x,y) \<in> r^* \<and> y \<in> NF r"
  from steps NF have py: "?p y" by simp
  hence pNF: "?p (some_NF r x)" unfolding some_NF_def 
    by (rule someI)
  from py have y: "(x,y) \<in> r^!" by auto
  from pNF have nf: "(x,some_NF r x) \<in> r^!" by auto
  from UNF[unfolded UNF_on_def] y nf show ?thesis by auto
qed

definition weak_diamond :: "'a ars \<Rightarrow> bool" ("w\<diamond>") where "w\<diamond> r \<equiv> (r\<inverse> O r) - Id \<subseteq> (r O r\<inverse>)"

lemma weak_diamond_imp_CR: assumes wd: "w\<diamond> r"
  shows "CR r"
proof (rule semi_confluence_imp_CR, rule)
  fix x y
  assume "(x,y) \<in> r^-1 O r^*"
  then obtain z where step: "(z,x) \<in> r" and steps: "(z,y) \<in> r^*" by auto
  from steps
  have "\<exists> u. (x,u) \<in> r^* \<and> (y,u) \<in> r^=" 
  proof (induct) 
    case base
    show ?case
      by (rule exI[of _ x], insert step, auto)
  next
    case (step y' y)
    from step(3) obtain u where xu: "(x,u) \<in> r^*" and y'u: "(y',u) \<in> r^=" by auto
    from y'u have "(y',u) \<in> r \<or> y' = u" by auto
    thus ?case
    proof
      assume y'u: "y' = u"
      with xu step(2) have xy: "(x,y) \<in> r^*" by auto
      show ?thesis 
        by (intro exI conjI, rule xy, simp)
    next
      assume "(y',u) \<in> r"
      with step(2) have uy: "(u,y) \<in> r^-1 O r" by auto
      show ?thesis
      proof (cases "u = y")
        case True
        show ?thesis
          by (intro exI conjI, rule xu, unfold True, simp)
      next
        case False
        with uy
          wd[unfolded weak_diamond_def] obtain u' where uu': "(u,u') \<in> r"
          and yu': "(y,u') \<in> r" by auto
        from xu uu' have xu: "(x,u') \<in> r^*" by auto
        show ?thesis
          by (intro exI conjI, rule xu, insert yu', auto)
      qed
    qed
  qed        
  thus "(x,y) \<in> r\<^sup>\<down>" by auto
qed

end