(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Ceta
imports
  Ceta_Verifier 
  "~~/src/HOL/Library/Code_Char_chr"
  Efficient_Sort
(*  "~~/src/HOL/Library/Code_Char_ord" clashes, but would be nice then one
   can remove code-setup for > and \<ge> below *)
  (*Efficient_Nat*)
begin



(* code_instance unit :: eq
  (Haskell -) *)

(* these axioms are required, since otherwise the code-generator is not able 
   to export these parser-functions *)
(* note that the correctness of certifyProof has already been proven without
   these axioms !! *)
(* RT: since I'm pretty sure, that all these functions are terminating:
       if someone has some spare time, go ahead and prove it *)
axiomatization where 
 parser_terminates:
  " (\<forall>x. plain_name_dom x) 
  \<and> (\<forall>x. term_dom x) 
  \<and> (\<forall>x. polynomial_dom x) 
  \<and> (\<forall>x. xml2trs_proof_dom x) 
  \<and> (\<forall>x. xml2reltrs_proof_dom x) 
  \<and> (\<forall>x. xml2dp_proof_dom x) 
  \<and> (\<forall>x. CPF_Parser.ctxt_dom x)
  \<and> (\<forall>x. xml2dp_disproof_dom x)
  \<and> (\<forall>x. xml2trs_disprof_dom x)
  \<and> (\<forall>x. xml2reltrs_disprof_dom x)
  \<and> (\<forall>x. arith_fun_dom x)"

termination CPF_Parser.plain_name by (simp add: parser_terminates)
termination CPF_Parser.term by (simp add: parser_terminates)
termination CPF_Parser.ctxt by (simp add: parser_terminates)
termination CPF_Parser.polynomial by (simp add: parser_terminates)
termination xml2dp_proof by (simp add: parser_terminates)
termination xml2trs_proof by (simp add: parser_terminates)
termination xml2reltrs_proof by (simp add: parser_terminates)
termination xml2reltrs_disproof by (simp add: parser_terminates)
termination xml2dp_disproof by (simp add: parser_terminates)
termination xml2trs_disproof by (simp add: parser_terminates)
termination CPF_Parser.arith_fun by (simp add: parser_terminates) 

(*
text {*
Setup the "show" type class as the standard type class
"Show" for Haskell.
*}
code_class "show"
  (Haskell "Show")

code_const shows_prec and shows_list and shows_string
  (Haskell "showsPrec" and "showList" and "showString")

text {*
Remove instances of "show" that are already present in Haskell.
*}
code_instance "*" :: "show" and char :: "show" and int :: "show" and list :: "show"
  (Haskell - and - and - and -)
*)

definition "termifyCR = (Code_Evaluation.term_of :: cert_result \<Rightarrow> Code_Evaluation.term)"
definition "termifyCP = (Code_Evaluation.term_of :: string + (string,label_type,string) cert_problem \<Rightarrow> Code_Evaluation.term)"

text {*
Don't use Haskell's "Show" type class, but rather invent a new name.
*}
code_reserved Haskell "Show"

code_const "op < :: char \<Rightarrow> char \<Rightarrow> bool"
  (SML "!((_ : char) < _)")
  (OCaml "!((_ : char) < _)")
  (Haskell infix 4 "<")
  (Scala infixl 4 "<")
  (Eval infixl 6 "<")

code_const "op \<le> :: char \<Rightarrow> char \<Rightarrow> bool"
  (SML "!((_ : char) <= _)")
  (OCaml "!((_ : char) <= _)")
  (Haskell infix 4 "<=")
  (Scala infixl 4 "<=")
  (Eval infixl 6 "<=")


(*
export_code certify_proof in Haskell module_name Ceta file "../generated/Haskell"
export_code certify_proof in OCaml module_name Ceta file "generated/OCaml/Ceta.ml"
export_code certify_proof in SML module_name Ceta file "generated/SML/Ceta.ML"
export_code certify_proof in Scala module_name Ceta file "generated/Scala/Ceta.scala"
export_code certify_proof parseCertProblem termifyCR termifyCP in Eval module_name EvalCeta file "generated/SML/EvalCeta.ML"
*)

end
