(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Matrix_Poly
imports Linear_Polynomial Matrix SN_Orders
begin


locale matrix_carrier = SN_one_mono_ordered_semiring_1 d gt
  for gt :: "'a :: {show,max_ordered_semiring_1} \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<succ>" 50) and d :: "'a"
begin

lemma mat_poly_order_carrier: assumes sd_n: "sd \<le> n" and sd_pos: "sd > 0" shows 
  "Linear_Polynomial.poly_order_carrier (mat n n) (mat0 n n) mat_plus (mat_mult n) mat_ge (mat_gt sd) True (mat1 n) (mat_default n) mat_max0 (\<lambda> x. True)"
  by (rule Linear_Polynomial.poly_order_carrier.intro, rule Linear_Polynomial.poly_carrier.intro,
    rule mat0, 
    rule mat_plus, simp+,
    rule mat_plus_comm,
    rule mat_plus_assoc, simp+,
    simp only: mat_plus_comm[of "mat0 n n"], rule plus_mat0, simp,
    rule mat_mult, simp+,
    rule mat_mult_assoc_n[symmetric], simp+,
    rule mat0_mult_left, simp, 
    rule mat0_mult_right, simp,
    rule mat_mult_plus_distrib_right, simp+,
    rule mat_mult_plus_distrib_left, simp+,
    rule mat_gt_compat[OF sd_n], simp+,
    rule mat_gt_compat2[OF sd_n], simp+,
    rule mat_plus_left_mono, simp+,
    rule mat_mult_left_mono, simp+,
    rule mat_mult_right_mono, simp+,
    rule mat_ge_refl, 
    rule mat_ge_trans, simp+,
    rule mat_gt_trans[OF sd_n], simp+,
    rule mat_plus_gt_left_mono[OF sd_n _ mat_ge_refl], simp+,
    rule Linear_Polynomial.poly_order_carrier_axioms.intro,
    simp add:  mat_sn[OF sd_n],
    rule mat_max0_id_pos, simp+,
    rule mat_max0_pos, simp+,
    rule mat_max0_mono, simp+,
    rule mat_max0_x, simp+,
    rule mat_max0, simp,
    rule mat1,
    rule mat1,
    rule mat1_mult_left, simp,
    rule mat1_mult_right, simp,
    rule mat1_ge_mat0,
    rule mat_default_ge_mat0,
    simp_all)
end

  
locale mono_matrix_carrier = SN_strict_mono_ordered_semiring_1 d gt + matrix_carrier gt d
  for gt :: "'a :: {show,max_ordered_semiring_1} \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<succ>" 50) and d :: 'a
begin

lemma mat_mono_poly_order_carrier: assumes sd_n: "sd \<le> n" and sd_pos: "sd > 0" shows 
  "mono_poly_order_carrier (mat n n) (mat0 n n) mat_plus (mat_mult n) mat_ge (mat_gt sd) True (mat1 n) (mat_default n) mat_max0 (\<lambda> x. True) (mat_mono sd)"
  by (rule mono_poly_order_carrier.intro, rule mat_poly_order_carrier[OF sd_n sd_pos],
    rule mono_poly_order_carrier_axioms.intro,
    simp,
    rule mat_mono[OF sd_n], 
    (simp_all)[6],
    rule mat_default_gt_mat0[OF sd_pos sd_n])

lemma linear_poly_order_carrier: "Linear_Polynomial.poly_order_carrier (\<lambda> x. True) 0 (op +) (op *) ge gt True 1 d max0 (\<lambda> x. True)"
proof(unfold_locales, simp_all add: field_simps SN max0_x max0_mono max0_pos max0_id_pos plus_left_mono times_left_mono times_right_mono one_ge_zero ge_refl default_ge_zero)
  fix s t u
  assume "s \<succeq> t" and "t \<succ> u"
  thus "s \<succ> u" by (rule compat)
next
  fix s t u
  assume "s \<succ> t" and "t \<succeq> u"
  thus "s \<succ> u" by (rule compat2)
next
  fix s t u
  assume "s \<succeq> t" and "t \<succeq> u"
  thus "s \<succeq> u" by (rule ge_trans)
next
  fix x y z u :: 'a
  assume gt: "x \<succ> y" 
  thus "x + z \<succ> y + z"
    by (rule plus_gt_left_mono)
next
  fix x y z :: 'a
  assume gt: "x \<succ> y" "y \<succ> z"
  thus "x \<succ> z"
    by (rule gt_trans)
qed
end

lemma (in mono_matrix_carrier) mono_poly_order_carrier: "mono_poly_order_carrier (\<lambda> x. True) 0 (op +) (op *) ge gt True 1 d max0 (\<lambda> x. True) mono"
proof -
  interpret Linear_Polynomial.poly_order_carrier "\<lambda> x. True" 0 "op +" "op *" ge gt True 1 d max0 "\<lambda> x. True"
    by (rule linear_poly_order_carrier)
  show ?thesis
    by (unfold_locales, auto simp: mono default_gt_zero)
qed


locale both_mono_matrix_carrier = SN_both_mono_ordered_semiring_1 d gt
  for gt :: "'a :: {show,ordered_semiring_1} \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<succ>" 50) and d :: 'a
begin

lemma mat_poly_order_carrier: assumes n_pos: "n > 0" shows 
  "Linear_Polynomial.poly_order_carrier (mat n n) (mat0 n n) mat_plus (mat_mult n) mat_ge mat_gt_arc False (mat1 n) (mat_default n) id (mat_arc_posI arc_pos)"
  by (rule Linear_Polynomial.poly_order_carrier.intro, rule Linear_Polynomial.poly_carrier.intro, 
    rule mat0, 
    rule mat_plus, simp+,
    rule mat_plus_comm,
    rule mat_plus_assoc, simp+,
    simp only: mat_plus_comm[of "mat0 n n"], rule plus_mat0, simp,
    rule mat_mult, simp+,
    rule mat_mult_assoc_n[symmetric], simp+,
    rule mat0_mult_left, simp, 
    rule mat0_mult_right, simp,
    rule mat_mult_plus_distrib_right, simp+,
    rule mat_mult_plus_distrib_left, simp+,
    rule mat_gt_arc_compat, simp+,
    rule mat_gt_arc_compat2, simp+,
    rule mat_plus_left_mono, simp+,
    rule mat_mult_left_mono, simp+,
    rule mat_mult_right_mono, simp+,
    rule mat_ge_refl, 
    rule mat_ge_trans, simp+, 
    rule mat_gt_arc_trans, simp+,
    rule mat_gt_arc_plus_mono, simp+,
    rule mat_gt_arc_mult_mono, simp+,
    rule mat0_leastI, simp+,
    rule mat0_leastII, simp+, 
    rule Linear_Polynomial.poly_order_carrier_axioms.intro,
    rule SN_subset[OF mat_gt_arc_SN[OF n_pos]], force,
    simp,
    simp add: mat0_leastIII,
    simp,
    simp add: mat_ge_refl,
    simp,
    rule mat1,
    rule mat1, 
    rule mat1_mult_left, simp,
    rule mat1_mult_right, simp,
    rule mat1_ge_mat0,
    rule mat_default_ge_mat0,
    rule mat_arc_pos_plus[OF n_pos], simp+,
    rule mat_arc_pos_mult[OF n_pos], simp+,
    rule mat_arc_pos_mat_default[OF n_pos]) 
end

lemma (in both_mono_matrix_carrier) linear_poly_order_carrier: "Linear_Polynomial.poly_order_carrier (\<lambda> x. True) 0 (op +) (op *) ge gt False 1 d id arc_pos"
proof(unfold_locales)
  fix s t u
  assume "s \<succ> t" and "t \<succeq> u"
  thus "s \<succ> u" by (rule compat2)
next
  fix s t u
  assume "s \<succeq> t" and "t \<succeq> u"
  thus "s \<succeq> u" by (rule ge_trans)
next
  fix s t u
  assume "s \<succ> t" and "t \<succ> u"
  thus "s \<succ> u" by (rule gt_trans)
qed (simp_all add: field_simps plus_left_mono times_left_mono times_right_mono zero_leastI SN zero_leastII zero_leastIII one_ge_zero ge_refl arc_pos_default arc_pos_plus arc_pos_mult compat plus_gt_both_mono times_gt_left_mono)
  
 
end
