(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Reduction_Pair
imports
  Term_Order
  QDP_Framework 
begin

fun
  nr_ichain ::
    "('f, 'v)trs \<times> ('f,'v)trs \<times> ('f,'v)trs \<Rightarrow> (nat \<Rightarrow> ('f, 'v) term) \<Rightarrow> (nat \<Rightarrow> ('f, 'v) term) \<Rightarrow>
     (nat \<Rightarrow> ('f, 'v) subst) \<Rightarrow> bool"
where
  "nr_ichain (P, Pw, R) s t \<sigma> = (
    (\<forall>i. (s i,t i) \<in> P \<union> Pw) \<and> 
    (INFM i. (s i,t i) \<in> P) \<and>
    (\<forall>i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (nrrstep R)^*))"

lemma ichain_imp_nr_ichain: 
  assumes ichain: "ichain (P,Pw,Q,{},R) s t \<sigma>"
  and nvar: "\<forall> (l,r) \<in> R. is_Fun l"
  and nvarP: "\<forall> (s,t) \<in> P \<union> Pw. is_Fun t"
  and ndef: "\<forall> (s,t) \<in> P \<union> Pw. \<not> defined R (the (root t))"
  shows "nr_ichain (P,Pw,R) s t \<sigma>"
proof -
  from ichain[unfolded ichain.simps] have 
    one: "\<forall>i. (s i,t i) \<in> P \<union> Pw"
    and two: "(INFM i. (s i, t i) \<in> P)"
    and three: "(\<forall>i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (qrstep Q R)^*)"
    by auto
  {
    fix i
    from three have steps: "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep R)^*" 
      using rtrancl_mono[OF qrstep_mono[OF subset_refl, of Q "{}" R]] by auto
    from one nvarP have t: "is_Fun (t i)" by auto
    hence nvart: "is_Fun (t i \<cdot> \<sigma> i)" by auto
    from one ndef have "\<not> defined R (the (root (t i)))" by auto
    with t have "\<not> defined R (the (root (t i \<cdot> \<sigma> i)))" by auto
    from rsteps_imp_nrrsteps[OF nvart this nvar steps]
    have "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (nrrstep R)^*" .
  }
  with one two show ?thesis by auto
qed

lemma nr_ichain_mono: assumes ichain: "nr_ichain (P,Pw,R) s t \<sigma>"
  and P: "P \<subseteq> P'"
  and Pw: "P \<union> Pw \<subseteq> P' \<union> Pw'"
  and R: "R \<subseteq> R'"
  shows "nr_ichain (P',Pw',R') s t \<sigma>"
proof -
  from ichain have
    one: "\<forall>i. (s i,t i) \<in> P \<union> Pw"
    and two: "(INFM i. (s i, t i) \<in> P)"
    and three: "(\<forall>i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (nrrstep R)^*)"
    by auto
  show ?thesis unfolding nr_ichain.simps 
  proof (intro conjI)
    show "\<forall>i. (s i,t i) \<in> P' \<union> Pw'" using one Pw by auto
  next
    show "(INFM i. (s i, t i) \<in> P')" using two P unfolding INFM_nat_le by auto
  next
    show "(\<forall>i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (nrrstep R')^*)"
      using rtrancl_mono[OF nrrstep_mono[OF R]] three by auto
  qed
qed

context root_redtriple
begin
lemma root_redtriple_nr_ichain_sound:
  assumes R: "R \<subseteq> NS"
  and P: "P \<subseteq> NST \<union> S"
  shows "\<not> nr_ichain (S \<inter> P, P - S, R) s t \<sigma>"
proof 
  assume chain': "nr_ichain (S \<inter> P, P - S, R) s t \<sigma>"
  have chain: "nr_ichain (S, NST \<union> S, NS) s t \<sigma>"
    by (rule nr_ichain_mono[OF chain'], insert R P, auto)
  let ?ss = "\<lambda> i. s i \<cdot> \<sigma> i"
  let ?ts = "\<lambda> i. t i \<cdot> \<sigma> i"
  let ?next = "\<lambda> i. (s (Suc i), t (Suc i))"
  {
    fix i
    from chain chain' have "(?ts i, ?ss (Suc i)) \<in> (nrrstep NS)^*" "(?ts i, ?ss (Suc i)) \<in> (nrrstep R)^*" by auto
  } note steps = this
  from chain have allP: "\<forall> i. (s i, t i) \<in> NST \<union> S" by (auto simp: ichain.simps)
  from chain have inf: "INFM i. (s i, t i) \<in> S" by auto
  have piece1: "\<forall> i. (?ts i, ?ts (Suc i)) \<in> S \<or> (?ts i, ?ts (Suc i)) \<in> (nrrstep NS \<union> NST)^* \<and> (?next i) \<notin> S"
  proof
    fix i
    show "(?ts i, ?ts (Suc i)) \<in> S \<or> (?ts i, ?ts (Suc i)) \<in> (nrrstep NS \<union> NST)^* \<and> (?next i) \<notin> S" 
    proof (cases "?next i \<in> S")
      case True
      hence "(?ss (Suc i), ?ts (Suc i)) \<in> S" using subst_S by (auto simp: subst_closed_def)
      from compat_NSTs[OF nrrsteps_imp_NST[OF R steps(2)] this]
      show ?thesis ..
    next
      case False
      with allP[THEN spec[of _ "Suc i"]]
      have one: "?next i \<in> NST" and two: "?next i \<notin> S" by auto
      from one have "(?ss (Suc i), ?ts (Suc i)) \<in> NST" using subst_NST by (auto simp: subst_closed_def)
      with steps(1)[of i] have steps: "(?ts i, ?ts (Suc i)) \<in> (nrrstep NS)^* O NST" by auto
      have "(?ts i, ?ts (Suc i)) \<in> (nrrstep NS \<union> NST)^*"
        by (rule set_mp[OF _ steps], regexp)
      with two show ?thesis by simp
    qed
  qed  
  hence infSeq: "\<forall> i. (?ts i, ?ts (Suc i)) \<in> (nrrstep NS \<union> NST)^* \<union> S" by auto
  from SN have "SN_on S {?ts 0}" by best
  from infSeq nrr.trCompat this have "\<exists> j. \<forall> i \<ge> j. (?ts i, ?ts (Suc i)) \<in> (nrrstep NS \<union> NST)^* - S" by (rule non_strict_ending)  
  from this obtain j where one: "\<forall> i \<ge> j. (?ts i, ?ts (Suc i)) \<in> (nrrstep NS \<union> NST)^* - S" ..
  with piece1 have ns: "\<forall> i \<ge> j. ?next i \<notin> S" by blast
  from inf[unfolded INFM_nat] obtain n where n: "n > j" and s: "(s n, t n) \<in> S" by auto
  from n obtain m where n: "n = Suc m" and m: "m \<ge> j" by (cases n, auto)
  from ns[THEN spec[of _ m]] s show False unfolding n using m by auto
qed

lemma root_redtriple_sound:
  assumes R: "R \<subseteq> NS"
  and P: "P \<subseteq> NST \<union> S"
  and nvar: "\<forall> (l,r) \<in> R. is_Fun l"
  and nvarP: "\<forall> (s,t) \<in> P. is_Fun t"
  and ndef: "\<forall> (s,t) \<in> P. \<not> defined R (the (root t))"
  shows "\<not> ichain (S \<inter> P, P - S, Q, {}, R) s t \<sigma>"
proof 
  assume ichain: "ichain (S \<inter> P, P - S, Q, {}, R) s t \<sigma>"
  have "nr_ichain (S \<inter> P, P - S, R) s t \<sigma>"
    by (rule ichain_imp_nr_ichain[OF ichain nvar], insert ndef nvarP, auto)
  with root_redtriple_nr_ichain_sound[OF R P]
  show False ..
qed
end


context redtriple
begin
lemma redtriple_sound:
  assumes R: "R \<subseteq> NS"
  and P: "P \<subseteq> NST \<union> S"
  shows "\<not> ichain (S \<inter> P, P - S, Q, {}, R) s t \<sigma>"
proof 
  assume chain: "ichain (S \<inter> P, P - S, Q, {}, R) s t \<sigma>"
  have chain: "ichain (S, NST \<union> S, {}, {}, NS) s t \<sigma>"
    by (rule ichain_mono[OF chain], insert R P, auto)
  let ?ss = "\<lambda> i. s i \<cdot> \<sigma> i"
  let ?ts = "\<lambda> i. t i \<cdot> \<sigma> i"
  let ?next = "\<lambda> i. (s (Suc i), t (Suc i))"
  from chain have allP: "\<forall> i. (s i, t i) \<in> NST \<union> S" by (auto simp: ichain.simps)
  from chain have steps: "\<forall> i. (?ts i, ?ss (Suc i)) \<in> (rstep NS)^*" 
    by (auto simp: ichain.simps)
  with rstep_NS_rtrancl have stepsNS: "\<And> i. (?ts i, ?ss (Suc i)) \<in> NS^*" by auto
  from chain have inf: "INFM i. (s i, t i) \<in> S" unfolding ichain_alternative by auto
  have piece1: "\<forall> i. (?ts i, ?ts (Suc i)) \<in> S \<or> (?ts i, ?ts (Suc i)) \<in> (NS \<union> NST)^* \<and> (?next i) \<notin> S"
  proof
    fix i
    show "(?ts i, ?ts (Suc i)) \<in> S \<or> (?ts i, ?ts (Suc i)) \<in> (NS \<union> NST)^* \<and> (?next i) \<notin> S" 
    proof (cases "?next i \<in> S")
      case True
      hence "(?ss (Suc i), ?ts (Suc i)) \<in> S" using subst_S by (auto simp: subst_closed_def)
      with stepsNS NS.trCompat show ?thesis by auto 
    next
      case False
      with stepsNS allP[THEN spec[of _ "Suc i"]]
      have one: "?next i \<in> NST" and two: "?next i \<notin> S" by auto
      from one have "(?ss (Suc i), ?ts (Suc i)) \<in> NST" using subst_NST by (auto simp: subst_closed_def)
      with stepsNS[of i] have steps: "(?ts i, ?ts (Suc i)) \<in> NS^* O NST" by auto
      have "(?ts i, ?ts (Suc i)) \<in> (NS \<union> NST)^*"
        by (rule set_mp[OF _ steps], regexp)
      with two show ?thesis by simp
    qed
  qed  
  hence infSeq: "\<forall> i. (?ts i, ?ts (Suc i)) \<in> (NS \<union> NST)^* \<union> S" by auto
  from SN have "SN_on S {?ts 0}" by best
  from infSeq both.trCompat this have "\<exists> j. \<forall> i \<ge> j. (?ts i, ?ts (Suc i)) \<in> (NS \<union> NST)^* - S" by (rule non_strict_ending)  
  from this obtain j where one: "\<forall> i \<ge> j. (?ts i, ?ts (Suc i)) \<in> (NS \<union> NST)^* - S" ..
  with piece1 have ns: "\<forall> i \<ge> j. ?next i \<notin> S" by blast
  from inf[unfolded INFM_nat] obtain n where n: "n > j" and s: "(s n, t n) \<in> S" by auto
  from n obtain m where n: "n = Suc m" and m: "m \<ge> j" by (cases n, auto)
  from ns[THEN spec[of _ m]] s show False unfolding n using m by auto
qed
end

fun redtriple_proc :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)dpp \<Rightarrow> ('f,'v)dpp \<Rightarrow> bool"
 where "redtriple_proc S NS NST Ps (P,Pw,Q,R,Rw) (P',Pw',Q',R',Rw') = (R \<union> Rw' \<subseteq> NS \<and> P \<union> Pw \<subseteq> NST \<union> S \<and> (P \<union> Pw) \<inter> Ps \<subseteq> S \<and> P' = P - Ps \<and> Pw' = Pw - Ps \<and> R' = R \<and> Rw' = Rw \<and> Q' = Q)"

fun root_redtriple_proc :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)dpp \<Rightarrow> ('f,'v)dpp \<Rightarrow> bool"
 where "root_redtriple_proc S NS NST Ps (P,Pw,Q,R,Rw) (P',Pw',Q',R',Rw') = (R \<union> Rw' \<subseteq> NS \<and> P \<union> Pw \<subseteq> NST \<union> S \<and> (P \<union> Pw) \<inter> Ps \<subseteq> S \<and> P' = P - Ps \<and> Pw' = Pw - Ps \<and> R' = R \<and> Rw' = Rw \<and> Q' = Q \<and> (\<forall> (s,t) \<in> P \<union> Pw. is_Fun t) \<and> (\<forall> (s,t) \<in> P \<union> Pw. \<not> defined (R \<union> Rw) (the (root t))))"


lemma (in root_redtriple) root_redtriple_proc_ci: "ci_proc (root_redtriple_proc S NS NST Ps)"
unfolding ci_proc_def
proof (intro allI impI)
  fix P Pw Q R Rw P' Pw' Q' R' Rw'
  assume proc: "root_redtriple_proc S NS NST Ps (P,Pw,Q,R,Rw) (P',Pw',Q',R',Rw')"  
  from proc have R: "R' = R" and Rw: "Rw' = Rw" and Q: "Q' = Q" and P: "P' = P - Ps" 
    and Pw: "Pw' = Pw - Ps" and Ps: "Ps \<inter> (P \<union> Pw) \<subseteq> S" by auto
  show "R' \<union> Rw' \<subseteq>ci R \<union> Rw \<and> Q = Q' \<and> (\<forall> s t \<sigma> . min_ichain (P, Pw, Q, R, Rw) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (P',Pw',Q',R', Rw') (shift s i) (shift t i) (shift \<sigma> i)))"  
  proof (intro conjI, unfold R Rw Q P Pw, simp add: ci_subset_refl, simp, intro allI impI)
    fix s t \<sigma>
    assume min: "min_ichain (P,Pw,Q,R,Rw) s t \<sigma>"
    from min_ichain_imp_var_cond[OF min] have varR: "\<forall> (l,r) \<in> R \<union> Rw. is_Fun l" by force
    from min have ichain: "ichain (P,Pw,Q,R,Rw) s t \<sigma>" by simp
    show "\<exists> i. ichain (P - Ps, Pw - Ps, Q, R, Rw) (shift s i) (shift t i) (shift \<sigma> i)"
    proof (rule ichain_split_P[OF ichain], rule)
      assume chain: "ichain (Ps \<inter> (P \<union> Pw), P \<union> Pw - Ps ,Q,{},R \<union> Rw) s t \<sigma>"
      have "ichain (S \<inter> (P \<union> Pw), P \<union> Pw - S ,Q,{},R \<union> Rw) s t \<sigma>"
        by (rule ichain_mono[OF chain], insert proc, auto)
      moreover have "\<not> ichain (S \<inter> (P \<union> Pw), P \<union> Pw - S ,Q,{},R \<union> Rw) s t \<sigma>"
        by (rule root_redtriple_sound, insert proc varR, auto)
      ultimately show False by simp
    qed
  qed
qed

context redtriple
begin
lemma redtriple_proc_ci: "ci_proc (redtriple_proc S NS NST Ps)"
unfolding ci_proc_def
proof (intro allI impI)
  fix P Pw Q R Rw P' Pw' Q' R' Rw'
  assume proc: "redtriple_proc S NS NST Ps (P,Pw,Q,R,Rw) (P',Pw',Q',R',Rw')"  
  from proc have R: "R' = R" and Rw: "Rw' = Rw" and Q: "Q' = Q" and P: "P' = P - Ps" 
    and Pw: "Pw' = Pw - Ps" and Ps: "Ps \<inter> (P \<union> Pw) \<subseteq> S" by auto
  show "R' \<union> Rw' \<subseteq>ci R \<union> Rw \<and> Q = Q' \<and> (\<forall> s t \<sigma> . min_ichain (P, Pw, Q, R, Rw) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (P',Pw',Q',R', Rw') (shift s i) (shift t i) (shift \<sigma> i)))"  
  proof (intro conjI, unfold R Rw Q P Pw, simp add: ci_subset_refl, simp, intro allI impI)
    fix s t \<sigma>
    assume min: "min_ichain (P,Pw,Q,R,Rw) s t \<sigma>"
    hence ichain: "ichain (P,Pw,Q,R,Rw) s t \<sigma>" by simp
    show "\<exists> i. ichain (P - Ps, Pw - Ps, Q, R, Rw) (shift s i) (shift t i) (shift \<sigma> i)"
    proof (rule ichain_split_P[OF ichain], rule)
      assume chain: "ichain (Ps \<inter> (P \<union> Pw), P \<union> Pw - Ps ,Q,{},R \<union> Rw) s t \<sigma>"
      have "ichain (S \<inter> (P \<union> Pw), P \<union> Pw - S ,Q,{},NS) s t \<sigma>"
        by (rule ichain_mono[OF chain], insert proc, auto)
      moreover have "\<not> ichain (S \<inter> (P \<union> Pw), P \<union> Pw - S ,Q,{},NS) s t \<sigma>"
        by (rule redtriple_sound, insert proc, auto)
      ultimately show False by simp
    qed
  qed
qed

lemma rel_SN:   
  shows "rel_SN (S,NS)"
  unfolding rel_SN_def rel_step.simps
  by (rule SN_subset[OF compatible_SN'[OF NS.compat SN]], regexp)

lemma rel_SN_rstep:
  assumes mono: "one_step_ctxt_closed S"
  shows "rel_SN (rstep S, rstep NS)" 
  by (rule rel_SN_mono[OF trs_inclusion[OF mono subst_S subset_refl] trs_inclusion[OF wm_NS subst_NS subset_refl] rel_SN])

lemma mono_redpair_sound_ichain: 
  assumes mono: "one_step_ctxt_closed S"
  and R: "R \<subseteq> NS \<union> S"
  and P: "P \<subseteq> NS \<union> S"
  shows "\<not> ichain (S \<inter> P, P - S, Q, S \<inter> R, R - S) s t \<sigma>"
  by (rule rel_SN_ichain, rule rel_SN_mono[OF qrstep_mono[of _ S _ "{}"] qrstep_mono[of _ NS _ "{}"]], insert R P rel_SN_rstep[OF mono], auto)

lemma mono_redpair_sound_ideriv: 
  assumes mono: "one_step_ctxt_closed S"
  and R: "R \<subseteq> NS \<union> S"
  shows "\<not> qrs_ideriv Q (S \<inter> R) (R - S) ts"
  using rel_SN_rstep[OF mono] qrs_ideriv_mono[of "S \<inter> R" S "R - S" NS Q "{}", of ts]
    R unfolding qrs_ideriv_def rel_SN_ideriv by auto
end

fun mono_redpair_proc :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)dpp \<Rightarrow> ('f,'v)dpp \<Rightarrow> bool"
 where "mono_redpair_proc S NS Ps Rs (P,Pw,Q,R,Rw) (P',Pw',Q',R',Rw') = (one_step_ctxt_closed S \<and> R \<union> Rw' \<subseteq> NS \<union> S \<and> P \<union> Pw \<subseteq> NS \<union> S \<and> (P \<union> Pw) \<inter> Ps \<subseteq> S \<and> (R \<union> Rw) \<inter> Rs \<subseteq> S \<and> P' = P - Ps \<and> Pw' = Pw - Ps \<and> R' = R - Rs \<and> Rw' = Rw - Rs \<and> Q' = Q)"

fun mono_redpair_tt :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)qreltrs \<Rightarrow> ('f,'v)qreltrs \<Rightarrow> bool"
 where "mono_redpair_tt S NS Rs (Q,R,Rw) (Q',R',Rw') = (one_step_ctxt_closed S \<and> R \<union> Rw' \<subseteq> NS \<union> S \<and> (R \<union> Rw) \<inter> Rs \<subseteq> S \<and> R' = R - Rs \<and> Rw' = Rw - Rs \<and> Q' = Q)"

context redtriple
begin
lemma mono_redpair_proc_ci: "ci_proc (mono_redpair_proc S NS Ps Rs)"
unfolding ci_proc_def
proof (intro allI impI)
  fix P Pw Q R Rw P' Pw' Q' R' Rw'
  assume proc: "mono_redpair_proc S NS Ps Rs (P,Pw,Q,R,Rw) (P',Pw',Q',R',Rw')"  
  from proc have mono: "one_step_ctxt_closed S" and  R: "R' = R - Rs" and Rw: "Rw' = Rw - Rs" and Q: "Q' = Q" and P: "P' = P - Ps"
    and Pw: "Pw' = Pw - Ps" by auto
  show "R' \<union> Rw' \<subseteq>ci R \<union> Rw \<and> Q = Q' \<and> (\<forall> s t \<sigma> . min_ichain (P, Pw, Q, R, Rw) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (P',Pw',Q',R', Rw') (shift s i) (shift t i) (shift \<sigma> i)))"  
  proof (intro conjI, unfold R Rw Q P Pw, rule ci_subsetI, blast, simp, intro allI impI)
    fix s t \<sigma>
    assume min: "min_ichain (P,Pw,Q,R,Rw) s t \<sigma>"
    hence ichain: "ichain (P,Pw,Q,R,Rw) s t \<sigma>" by simp
    show "\<exists> i. ichain (P - Ps, Pw - Ps, Q, R - Rs, Rw - Rs) (shift s i) (shift t i) (shift \<sigma> i)"
    proof (rule ichain_split[OF ichain], rule)
      assume chain: "ichain (Ps \<inter> (P \<union> Pw), P \<union> Pw - Ps ,Q,Rs \<inter> (R \<union> Rw) ,R \<union> Rw - Rs) s t \<sigma>"
      have "ichain (S \<inter> (P \<union> Pw), P \<union> Pw - S ,Q, S \<inter> (R \<union> Rw), R \<union> Rw - S) s t \<sigma>"
        by (rule ichain_mono[OF chain], insert proc, auto)
      moreover have "\<not> ichain (S \<inter> (P \<union> Pw), P \<union> Pw - S ,Q, S \<inter> (R \<union> Rw),R \<union> Rw - S) s t \<sigma>"
        by (rule mono_redpair_sound_ichain[OF mono], insert proc, auto)
      ultimately show False by simp
    qed
  qed
qed

lemma mono_redpair_tt: "rel_tt (mono_redpair_tt S NS Rs)"
unfolding rel_tt_def
proof (intro allI impI)
  fix Q R Rw Q' R' Rw' ts
  assume proc: "mono_redpair_tt S NS Rs (Q,R,Rw) (Q',R',Rw')" 
    and ideriv: "qrs_ideriv Q R Rw ts"
  from proc have mono: "one_step_ctxt_closed S" and  R: "R' = R - Rs" and Rw: "Rw' = Rw - Rs" and Q: "Q' = Q" and Rns: "R \<union> Rw \<subseteq> NS \<union> S" by auto
  from qrs_ideriv_split[OF ideriv mono_redpair_sound_ideriv[OF mono Rns]]
  obtain i where ideriv: "qrs_ideriv Q (R - S) (Rw - S) (shift ts i)" ..
  have "qrs_ideriv Q (R - Rs) (Rw - Rs) (shift ts i)"
    by (rule qrs_ideriv_mono[OF _ _ _ ideriv], insert proc, auto)
  thus "Ex (qrs_ideriv Q' R' Rw')" unfolding R Rw Q by auto
qed

lemma manna_ness:
  assumes mono: "one_step_ctxt_closed S"
  and strict: "R \<subseteq> S"
  shows "SN (rstep R)"
proof -
  from mono strict subst_S trs_inclusion have "(rstep R) \<subseteq> S" by auto
  from SN_subset[OF SN this] show "SN (rstep R)"  .
qed
end

end