(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Rewriting
imports QDP_Framework
  Innermost_Usable_Rules
  Critical_Pairs
begin


context R_Q_U_ecap
begin

lemma rewriting_sound: 
  assumes rewrite: "(t,t') \<in> rstep_r_p_s R (l,r) p \<mu>"
  and usable: "is_ur_closed_term_mv {s} (t |_ p)"
  and var_cond: "vars_term (t |_ p) \<subseteq> vars_term s"
  and var_cond_2: "vars_term r \<subseteq> vars_term l"
  and sNF: "s \<cdot> \<sigma> \<in> NFQ"
  and unf: "UNF (qrstep Q U)"
  and steps: "(t \<cdot> \<sigma>, u) \<in> (qrstep Q R)^*"
  and crit_pairs: "\<And> b s t. (b,s,t) \<in> critical_pairs {(l,r)} U \<Longrightarrow> s = t"
  and subset: "U \<subseteq> R"
  and uNF: "u \<in> NFQ"
  and wf_trs_or_SN: "wf_trs R \<or> SN_on (qrstep Q R) {t \<cdot> \<sigma>}"
  and U: "\<And> l r. (l,r) \<in> U \<Longrightarrow> is_ur_closed_term_mv (set (args l)) r"  
  shows "(t' \<cdot> \<sigma>, u) \<in> (qrstep Q R)^* \<and> (SN_on (qrstep Q R) {t \<cdot> \<sigma>} \<longrightarrow> SN_on (qrstep Q R) {t' \<cdot> \<sigma>})"
proof -
  let ?QR = "qrstep Q R"
  let ?QU = "qrstep Q U"
  let ?QUR = "qrstep Q (R \<inter> U)"
  have "NFQ \<subseteq> NF_trs R" by (rule NF_Q_R)
  also have "... \<subseteq> NF_trs U" by (rule NF_trs_mono[OF subset])
  finally have NF_Q_U: "NFQ \<subseteq> NF_trs U" . 
  from rewrite[unfolded rstep_r_p_s_def'] 
  have p: "p \<in> poss t" and lr: "(l,r) \<in> R" and subt: "t |_ p = l \<cdot> \<mu>"
    and t': "t' = replace_at t p (r \<cdot> \<mu>)" by auto
  from p have pt\<sigma>: "p \<in> poss (t \<cdot> \<sigma>)" by auto
  from normalize_subterm_qrsteps[OF pt\<sigma> steps uNF] obtain w
    where steps1: "(t \<cdot> \<sigma> |_ p, w) \<in> ?QR^*" and w: "w \<in> NFQ" and steps2: "(replace_at (t \<cdot> \<sigma>) p w, u) \<in> ?QR^*" by auto
  from steps1 p have steps1: "(t |_ p \<cdot> \<sigma>, w) \<in> ?QR^*" by auto
  from subt_at_imp_supteq[OF p] have ttp: "t \<unrhd> t |_ p" .
  {
    fix x
    assume "x \<in> vars_term (t |_ p)"
    with var_cond have "x \<in> vars_term s" by auto
    hence "Var x \<unlhd> s" by auto
    hence "Var x \<cdot> \<sigma> \<unlhd> s \<cdot> \<sigma>" by (rule supteq_stable)
    from NF_subterm[OF sNF this]
    have "\<sigma> x \<in> NFQ" by simp
  } 
  hence \<sigma>tp: "\<sigma> ` vars_term (t |_ p) \<subseteq> NFQ" by auto
  from wf_trs_or_SN have wf: "wf_trs R \<or> SN_on ?QR {t |_ p \<cdot> \<sigma>}"
    using SN_imp_SN_subt_gen[OF ctxt_closed_qrstep, of Q R "t \<cdot> \<sigma>", OF _ supteq_stable[OF ttp]] by auto
  let ?\<mu>\<sigma> = "\<mu> \<circ>s \<sigma>"
  have steps1: "(t |_ p \<cdot> \<sigma>, w) \<in> ?QUR^*"
    by (rule is_ur_closed_term[OF steps1 \<sigma>tp _ U subset_refl wf usable], insert sNF, auto)
  with subt have steps1: "(l \<cdot> ?\<mu>\<sigma>, w) \<in> ?QUR^*" by auto
  obtain \<tau> where \<tau>: "\<tau> \<equiv> \<lambda> x. some_NF (qrstep Q U) (?\<mu>\<sigma> x)" by auto
  have steps1': "(l \<cdot> ?\<mu>\<sigma>, w) \<in> ?QU^*"
    by (rule set_mp[OF _ steps1], rule rtrancl_mono, rule qrstep_mono, auto)
  from normalize_subst_qrsteps_inn[OF unf _ steps1' w NF_Q_U]
  have lx: "\<And> x. x \<in> vars_term l \<Longrightarrow> (?\<mu>\<sigma> x, \<tau> x) \<in> ?QU^*"
    "\<And> x. x \<in> vars_term l \<Longrightarrow> \<tau> x \<in> NFQ"
    and steps: "(l \<cdot> \<tau>, w) \<in> ?QU^*" unfolding \<tau> by blast+
  {
    fix x
    assume "x \<in> vars_term l"
    from lx(2)[OF this] NF_Q_R
    have "\<tau> x \<in> NF_trs R" by auto
  } note \<tau>NF = this
  from lr have "(l \<cdot> \<tau>, r \<cdot> \<tau>) \<in> rstep R" by auto
  hence "l \<cdot> \<tau> \<notin> NF_trs R" by auto
  with NF_Q_R have "l \<cdot> \<tau> \<notin> NFQ" by auto
  with w have lw: "l \<cdot> \<tau> \<noteq> w" by auto
  from rtrancl_imp_rel_pow[OF steps] obtain n where n: "(l \<cdot> \<tau>, w) \<in> ?QU ^^ n" by auto
  with lw obtain n where "(l \<cdot> \<tau>, w) \<in> ?QU ^^ (Suc n)" by (cases n, auto)
  from rel_pow_Suc_D2[OF this] obtain v where step: "(l \<cdot> \<tau>, v) \<in> ?QU" and steps: "(v, w) \<in> ?QU ^^ n" by auto
  from rel_pow_imp_rtrancl[OF steps] have steps: "(v,w) \<in> ?QU^*" by auto
  {
    from step obtain l' r' p \<sigma> where step: "(l \<cdot> \<tau>,v) \<in> qrstep_r_p_s Q U (l',r') p \<sigma>" unfolding qrstep_qrstep_r_p_s_conv by auto
    obtain D where D: "D = ctxt_of_pos_term p (l \<cdot> \<tau>)" by auto
    from step have p: "p \<in> poss (l \<cdot> \<tau>)" and
      NF': "\<forall> u \<lhd> l' \<cdot> \<sigma>. u \<in> NFQ"
      and lr': "(l',r') \<in> U"
      and l\<tau>p: "l \<cdot> \<tau> |_ p = l' \<cdot> \<sigma>"
      and v: "v = D\<langle>r' \<cdot> \<sigma>\<rangle>" unfolding qrstep_r_p_s_def D by auto
    from ctxt_supt_id[OF p] have "l \<cdot> \<tau> = D\<langle>l \<cdot> \<tau> |_ p\<rangle>" unfolding D ..
    with l\<tau>p have l\<tau>: "l \<cdot> \<tau> = D\<langle>l' \<cdot> \<sigma>\<rangle>" by simp
    have "v = r \<cdot> \<tau>"
    proof (cases "p \<in> poss l \<and> is_Fun (l |_ p)")
      case False      
      from pos_into_subst[OF l\<tau> p[unfolded l\<tau>] this]
      obtain q q' x where pq: "p = q <#> q'" and q: "q \<in> poss l" and x: "l |_ q = Var x" by auto
      from q have q\<tau>: "q \<in> poss (l \<cdot> \<tau>)" by auto
      from l\<tau>p[unfolded pq subt_at_append[OF q\<tau>] subt_at_subst[OF q]]
      have id: "\<tau> x |_ q' = l' \<cdot> \<sigma>" unfolding x by auto
      from p[unfolded pq poss_append_poss subt_at_subst[OF q] x] have q': "q' \<in> poss (\<tau> x)" by auto
      from q' have "\<tau> x \<unrhd> \<tau> x |_ q'" by (rule subt_at_imp_supteq)
      with id have subt: "\<tau> x \<unrhd> l' \<cdot> \<sigma>" by auto
      from q have "l \<unrhd> l |_ q" by (rule subt_at_imp_supteq)
      from this[unfolded x] have "x \<in> vars_term l" by (rule var_subt_imp_subt_in_vars)
      from \<tau>NF[OF this] have NF: "\<tau> x \<in> NF_trs R" .
      from lr' subset have "(l',r') \<in> R" by auto
      from lhs_notin_NF_rstep[OF this]
        NF_instance[OF NF_subterm[OF NF subt]] show ?thesis by auto
    next
      case True
      hence p: "p \<in> poss l" and isfun: "is_Fun l |_ p" by auto
      from l\<tau>p have id: "l |_ p \<cdot> \<tau> = l' \<cdot> \<sigma>" unfolding subt_at_subst[OF p] .
      from ctxt_supt_id[OF p] obtain E where E': "E = ctxt_of_pos_term p l" and E: "E\<langle>l |_ p\<rangle> = l" by auto
      have lr: "(l,r) \<in> {(l,r)}" by auto
      from mgu_var_disjoint_string_complete[OF id]
      obtain \<mu>1 \<mu>2 \<mu>3 where mgu: "mgu_var_disjoint_string (l |_ p) l' = Some (\<mu>1,\<mu>2)" 
        and \<tau>: "\<tau> = \<mu>1 \<circ>s \<mu>3" and \<sigma>: "\<sigma> = \<mu>2 \<circ>s \<mu>3" by auto
      from crit_pairs[OF critical_pairsI[OF lr lr' E[symmetric] isfun mgu refl refl refl]]
      have id: "r \<cdot> \<mu>1 = (E \<cdot>c \<mu>1)\<langle>r' \<cdot> \<mu>2\<rangle>" .
      have id2: "D = E \<cdot>c \<mu>1 \<cdot>c \<mu>3" unfolding D E'
        unfolding ctxt_of_pos_term_subst[OF p]  \<tau> by simp
      show "v = r \<cdot> \<tau>" 
        unfolding v \<sigma> \<tau> subst_compose_apply_term_distrib id
        subst_apply_term_ctxt_apply_distrib id2 by simp
    qed
  } 
  hence vr\<tau>: "v = r \<cdot> \<tau>" .
  have lsteps: "(l \<cdot> ?\<mu>\<sigma>, l \<cdot> \<tau>) \<in> ?QU^*"
    by (rule all_ctxt_closed_subst_step[OF all_ctxt_closed_qrstep lx(1)])
  with step[unfolded vr\<tau>] have lsteps: "(l \<cdot> \<mu> \<cdot> \<sigma>, r \<cdot> \<tau>) \<in> ?QU^*" by auto
  from var_cond_2 have var_cond: "\<And> x. x \<in> vars_term r \<Longrightarrow> x \<in> vars_term l" by auto
  have rsteps: "(r \<cdot> ?\<mu>\<sigma>, r \<cdot> \<tau>) \<in> ?QU^*"
    by (rule all_ctxt_closed_subst_step[OF all_ctxt_closed_qrstep lx(1)[OF var_cond]])
  from rtrancl_mono[OF qrstep_mono[OF subset subset_refl]] have subsetUR: "?QU^* \<subseteq> ?QR^*" by simp
  from rsteps steps[unfolded vr\<tau>] have "(r \<cdot> \<mu> \<cdot> \<sigma>, w) \<in> ?QU^*" by auto
  with subsetUR have steps: "(r \<cdot> \<mu> \<cdot> \<sigma>, w) \<in> ?QR^*" by auto
  have t'\<sigma>: "t' \<cdot> \<sigma> = replace_at (t \<cdot> \<sigma>) p (r \<cdot> \<mu> \<cdot> \<sigma>)" unfolding t' ctxt_of_pos_term_subst[OF p] by simp
  have first_part: "(t' \<cdot> \<sigma>, u) \<in> ?QR^*" unfolding t'\<sigma> 
    by (rule rtrancl_trans[OF _ steps2], 
      rule one_step_ctxt_closed_ctxt[OF rtrancl_one_step_ctxt_closed[OF one_step_ctxt_closed_qrstep] steps])
  show ?thesis
  proof (rule conjI[OF first_part], intro impI)
    assume SN: "SN_on ?QR {t \<cdot> \<sigma>}"
    with SN_imp_SN_subt_gen[OF ctxt_closed_qrstep, of Q R "t \<cdot> \<sigma>", OF _ supteq_stable[OF ttp]] have SN_l: "SN_on ?QR {l \<cdot> \<mu> \<cdot> \<sigma>}" unfolding subt by auto 
    {
      fix x
      assume x: "x \<in> vars_term r"
      have subt_lx: "l \<cdot> \<mu> \<unrhd> Var x \<cdot> \<mu>" 
        by (rule supteq_stable, insert var_cond[OF x], auto)
    } note subt_lx = this
    {
      fix x
      assume x: "x \<in> vars_term r"
      from SN_imp_SN_subt_gen[OF ctxt_closed_qrstep, of Q R "l \<cdot> \<mu> \<cdot> \<sigma>", OF SN_l supteq_stable[OF subt_lx[OF x]]]
      have "SN_on ?QR {Var x \<cdot> \<mu> \<cdot> \<sigma>}" .
    } note SN_x = this
    {
      fix x u
      assume x: "x \<in> vars_term r" and steps: "(?\<mu>\<sigma> x, u) \<in> ?QR^*"
      from steps have steps: "(Var x \<cdot> \<mu> \<cdot> \<sigma>, u) \<in> ?QR^*" unfolding subst_compose_def by auto
      from \<sigma>tp[unfolded subt vars_term_subst] var_cond[OF x] have NF: "\<sigma> ` vars_term (Var x \<cdot> \<mu>) \<subseteq> NFQ" by auto
      have usable: "is_ur_closed_term_mv {s} (Var x \<cdot> \<mu>)"
        by (rule is_ur_closed_mv_subt[OF subt_lx[OF x] usable[unfolded subt]]) 
      from SN_x[OF x] have wf: "wf_trs R \<or> SN_on ?QR {Var x \<cdot> \<mu> \<cdot> \<sigma>}" ..
      from is_ur_closed_term[OF steps NF _ U subset_refl wf usable] sNF
      have "(Var x \<cdot> \<mu> \<cdot> \<sigma>, u) \<in> (qrstep Q (R \<inter> U))^*" by simp
      hence "(Var x \<cdot> ?\<mu>\<sigma>, u) \<in> ?QU^*"
        using rtrancl_mono[OF qrstep_mono[of "(R \<inter> U)" U Q, OF _ subset_refl]] by auto
      hence "(?\<mu>\<sigma> x, u) \<in> ?QU^*" unfolding subst_compose_def by auto
    } note x_R_U = this
    show "SN_on ?QR {t' \<cdot> \<sigma>}" unfolding t'\<sigma>
    proof (rule normalize_subterm_SN)
      show "SN_on ?QR {r \<cdot> \<mu> \<cdot> \<sigma>}" 
      proof (rule ccontr)
        assume "\<not> SN_on ?QR {r \<cdot> \<mu> \<cdot> \<sigma>}"
        hence nSN: "\<not> SN_on ?QR {r \<cdot> ?\<mu>\<sigma>}" by simp
        from SN_x have SN_x: "\<And> x. x \<in> vars_term r \<Longrightarrow> SN_on ?QR {?\<mu>\<sigma> x}" 
          unfolding subst_compose_def by auto
        from normalize_subst_qrsteps_inn_infinite[OF unf x_R_U nSN SN_x NF_Q_U]
        have nSN: "\<not> SN_on ?QR {r \<cdot> \<tau>}" unfolding \<tau> by blast+
        from steps_preserve_SN_on[OF _ SN_l] lsteps subsetUR have "SN_on ?QR {r \<cdot> \<tau>}" by auto
        with nSN show False by simp
      qed
    next
      fix v
      assume rv: "(r \<cdot> \<mu> \<cdot> \<sigma>, v) \<in> ?QR^*" and v: "v \<in> NFQ"
      hence rv': "(r \<cdot> ?\<mu>\<sigma>,v) \<in> ?QR^*" by auto
      have r\<tau>v: "(r \<cdot> \<tau>, v) \<in> ?QR^*" unfolding \<tau>
        using normalize_subst_qrsteps_inn[OF unf x_R_U rv' v NF_Q_U] by auto
      with lsteps subsetUR have lsteps: "(l \<cdot> \<mu> \<cdot> \<sigma>, v) \<in> ?QR^*" by auto
      have "t \<cdot> \<sigma> = replace_at (t \<cdot> \<sigma>) p (t \<cdot> \<sigma> |_p)"
        by (rule ctxt_supt_id[symmetric], rule pt\<sigma>)
      also have "... = replace_at (t \<cdot> \<sigma>) p (t |_p \<cdot> \<sigma>)"
        using p by auto
      also have "... = replace_at (t \<cdot> \<sigma>) p (l \<cdot> \<mu> \<cdot> \<sigma>)" unfolding subt ..
      also have "(..., replace_at (t \<cdot> \<sigma>) p v) \<in> ?QR^*"
        by (rule one_step_ctxt_closed_ctxt[OF rtrancl_one_step_ctxt_closed[OF one_step_ctxt_closed_qrstep] lsteps])
      finally have "(t \<cdot> \<sigma>, replace_at (t \<cdot> \<sigma>) p v) \<in> ?QR^*" .
      from steps_preserve_SN_on[OF this SN] show "SN_on ?QR {replace_at (t \<cdot> \<sigma>) p v}" .
    next
      show "SN_on ?QR (set (ctxt_to_term_list (ctxt_of_pos_term p (t \<cdot> \<sigma>))))"
        by (rule SN_ctxt_apply_imp_SN_ctxt_to_term_list_gen[OF ctxt_closed_qrstep, where t = "(t \<cdot> \<sigma>) |_ p"], unfold ctxt_supt_id[OF `p \<in> poss (t \<cdot> \<sigma>)`], rule SN)
    qed
  qed
qed
   

lemma rewriting_proc: assumes fin: "finite_dpp (replace (s1,t1) {(s1,t1')} P,replace (s1,t1) {(s1,t1')} Pw,Q,Rs,Rw)" (is "finite_dpp (?P,?Pw,_,_,_)")
  and rewrite: "(t1,t1') \<in> rstep_r_p_s R (l,r) p \<mu>"
  and usable: "is_ur_closed_term_mv {s1} (t1 |_ p)"
  and var_cond: "vars_term (t1 |_ p) \<subseteq> vars_term s1"
  and var_cond_2: "vars_term r \<subseteq> vars_term l"
  and unf: "UNF (qrstep Q U)"
  and strict: "(s1,t1) \<in> P \<or> Rs = {}"
  and crit_pairs: "\<And> b s t. (b,s,t) \<in> critical_pairs {(l,r)} U \<Longrightarrow> s = t"
  and subset: "U \<subseteq> R" 
  and U: "\<And> l r. (l,r) \<in> U \<Longrightarrow> is_ur_closed_term_mv (set (args l)) r"  
  and R: "Rs \<union> Rw = R"
  shows "finite_dpp (P, Pw, Q, Rs, Rw)"
proof (rule ccontr)
  assume "\<not> ?thesis"
  then obtain s t \<sigma> where chain: "min_ichain (P,Pw,Q,Rs,Rw) s t \<sigma>" 
    unfolding finite_dpp_def by auto
  from chain have min: "minimal_cond Q R s t \<sigma>" unfolding min_ichain.simps R by auto     
  let ?t = "\<lambda> i. if (s i,t i) = (s1,t1) then t1' else t i"
  let ?R = "qrstep Q R"
  let ?Rb = "qrstep Q (Rs \<union> Rw)"
  let ?Rs = "qrstep Q Rs"
  from chain have chain: "ichain (P,Pw,Q,Rs,Rw) s t \<sigma>" unfolding min_ichain.simps by simp
  note chain = chain[unfolded ichain.simps]
  from chain have inP: "\<And> i. (s i, t i) \<in> P \<union> Pw" by auto
  from chain have NF: "\<And> i. s i \<cdot> \<sigma> i \<in> NFQ" by auto
  from chain have steps: "\<And> i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?R^*" unfolding R by auto
  from min have SN: "\<And> i. SN_on ?R {t i \<cdot> \<sigma> i}" unfolding minimal_cond_def by auto
  have "min_ichain (?P,?Pw,Q,Rs,Rw) s ?t \<sigma>"
    unfolding min_ichain.simps
  proof
    show "minimal_cond Q (Rs \<union> Rw) s ?t \<sigma>" unfolding R minimal_cond_def 
    proof
      fix i
      show "SN_on ?R {?t i \<cdot> \<sigma> i}"
      proof (cases "(s i, t i) = (s1,t1)")
        case False
        with SN[of i] show ?thesis by auto
      next
        case True
        hence id: "?t i = t1'" "s i = s1" "t i = t1" by auto
        with SN[of i] have "SN_on ?R {t1 \<cdot> \<sigma> i}" by auto
        with rewriting_sound[OF rewrite usable var_cond var_cond_2 NF[of i, unfolded id] unf steps[of i, unfolded id] crit_pairs subset NF _ U] 
        show ?thesis unfolding id by simp
      qed
    qed
  next
    show "ichain (?P,?Pw,Q,Rs,Rw) s ?t \<sigma>"
      unfolding ichain.simps
    proof (intro conjI allI)
      fix i
      show "(s i, ?t i) \<in> ?P \<union> ?Pw" using inP[of i]
        by (cases "(s i, t i) = (s1,t1)", auto simp: replace_def)
    next        
      fix i
      show "s i \<cdot> \<sigma> i \<in> NFQ" by (rule NF)
    next
      fix i
      have "(?t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?R^*"
      proof (cases "(s i, t i) = (s1,t1)")
        case False
        with steps[of i] show ?thesis by auto
      next
        case True
        hence id: "?t i = t1'" "s i = s1" "t i = t1" by auto
        from rewriting_sound[OF rewrite usable var_cond var_cond_2 NF[of i, unfolded id] unf steps[of i, unfolded id] crit_pairs subset NF _ U] 
        show ?thesis unfolding id using SN[of i, unfolded id] by simp
      qed
      thus "(?t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?Rb^*" unfolding R .
    next
      from chain have disj: "(INFM i. (s i, t i) \<in> P) \<or> (INFM i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?Rb^* O ?Rs O ?Rb^*)" by simp
      show "(INFM i. (s i, ?t i) \<in> ?P) \<or> (INFM i. (?t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?Rb^* O ?Rs O ?Rb^*)" unfolding
        INFM_disj_distrib[symmetric] unfolding INFM_nat_le
      proof
        fix i
        from disj[unfolded INFM_nat_le] obtain j where j: "j \<ge> i" and 
          disj: "(s j, t j) \<in> P \<or> (t j \<cdot> \<sigma> j, s (Suc j) \<cdot> \<sigma> (Suc j)) \<in> ?Rb^* O ?Rs O ?Rb^*" by blast
        show "\<exists> j \<ge> i. (s j, ?t j) \<in> ?P \<or> (?t j \<cdot> \<sigma> j, s (Suc j) \<cdot> \<sigma> (Suc j)) \<in> ?Rb^* O ?Rs O ?Rb^*"
        proof (intro exI conjI, rule j)
          show "(s j, ?t j) \<in> ?P \<or> (?t j \<cdot> \<sigma> j, s (Suc j) \<cdot> \<sigma> (Suc j)) \<in> ?Rb^* O ?Rs O ?Rb^*"
            by (cases "(s j, t j) = (s1,t1)", insert strict disj, auto simp: replace_def)
        qed
      qed
    qed
  qed    
  with fin show False unfolding finite_dpp_def by auto
qed

end      
  
end