(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Signature_Extension
imports
  DP_Transformation
begin

text {* An auxiliary function to stack contexts (and substitutions) in order to reconstruct
DP steps in the original TRS. *}
fun
  D :: "(nat \<Rightarrow> ('f, 'v) ctxt) \<Rightarrow> (nat \<Rightarrow> ('f, 'v) subst) \<Rightarrow> nat \<Rightarrow> ('f, 'v) ctxt"
where
  "D C \<sigma> 0 = \<box>"
| "D C \<sigma> (Suc i) = (D C \<sigma> i) \<circ>c ((C i) \<cdot>c (\<sigma> i))"


lemma D_subst_wf:
  fixes F::"'f sig" and f::"('g, 'v) subst \<Rightarrow> ('f, 'v) subst"
  assumes \<sigma>: "\<And> i x. funas_term (\<sigma> i x) \<subseteq> F"
  assumes "\<forall>i. funas_ctxt (C i) \<subseteq> F" shows "funas_ctxt (D C \<sigma> i) \<subseteq> F"
proof (induct i)
  case 0 show ?case by auto
next
  case (Suc i)
  obtain s where s: "\<sigma> i = s" by auto
  from assms have Ci: "funas_ctxt (C i) \<subseteq> F" by simp
  have Ci: "funas_ctxt (C i \<cdot>c \<sigma> i) \<subseteq> F" using Ci \<sigma> by auto
  show ?case unfolding D.simps using Suc Ci by simp
qed

definition
  sharp_sig :: "('f \<Rightarrow> 'f) \<Rightarrow> 'f sig \<Rightarrow> 'f sig"
where
  "sharp_sig shp F \<equiv> F \<union> {(shp f, n) | f n. (f, n) \<in> F}"

lemma lhs_DP_wf:
  fixes R::"('f, 'v) trs" and shp :: "'f \<Rightarrow> 'f"
  defines "F \<equiv> sharp_sig shp (funas_trs R)"
  assumes DP: "(s,t) \<in> DP shp R" shows "funas_term s \<subseteq> F"
proof -
  have subset: "funas_trs(DP shp R) \<subseteq> F" using funas_DP_subset[of shp]
    unfolding F_def sharp_sig_def by auto
  have "funas_term s \<subseteq> funas_trs(DP shp R)" using DP unfolding funas_trs_def funas_rule_def by auto
  with subset show ?thesis by auto
qed



lemma the_inv_id: "the_inv id = id"
proof (intro ext)
  fix x
  have "the_inv id x = the_inv id (id x)" by simp
  also have "\<dots> = x" by (rule the_inv_f_f, unfold inj_on_def, simp)
  finally show "the_inv id x = id x" by simp
qed

lemma [simp]: "sharp_term id t = t" by (cases t, auto)
lemma [simp]: "sharp_term (the_inv id) t = t" 
proof (cases t, simp)
  case (Fun f ss)
  show ?thesis unfolding Fun  
    by (simp add: the_inv_id)
qed

lemma DP_not_is_Var_t: "(s,t) \<in> DP shp R \<Longrightarrow> is_Fun t"
  unfolding DP_def by auto

lemma DP_not_is_Var_s: "wf_trs R \<Longrightarrow> (s,t) \<in> DP shp R \<Longrightarrow> is_Fun s"
  unfolding DP_def wf_trs_def by force

lemma sharp_term_subst_distr_2: "is_Fun t \<Longrightarrow> sharp_term shp t \<cdot> \<sigma> = sharp_term shp (t \<cdot> \<sigma>)"
  using sharp_term_subst_distr by (cases t, auto)

lemma sharp_term_clean_term_cancel[simp]:
  assumes wf: "funas_term t \<subseteq> F"
  shows "sharp_term shp (clean_term F t) = sharp_term shp t"
proof (cases t)
  case (Var x) show ?thesis by (simp add: Var)
next
  case (Fun f ts)
  from wf[unfolded Fun]
    have ts: "map (clean_term F) ts = ts"
    using map_eq_conv[of "clean_term F" ts "id"] clean_term_ident[of _ F]
    by (cases) (auto simp: o_def)
  from wf[unfolded Fun] have "(f, length ts) \<in> F" by simp
  thus ?thesis by (simp add: ts Fun)
qed

lemma left_DP_is_sharped_term:
  assumes "(s, t) \<in> DP shp R"
  shows "\<exists>u \<in> lhss R. s = sharp_term shp u"
  using assms unfolding DP_def by force

(* TODO: perhaps specialize to arbitrary shp-functions instead of id *)
theorem clean_ichain_imp_not_SN_wfrstep:
  fixes R::"('f, 'v) trs" and shp :: "'f \<Rightarrow> 'f"
  defines F: "F \<equiv> sharp_sig shp (funas_trs R)"
  assumes chain: "ichain (DP shp R, {}, Q, {}, R) s t \<sigma>"
    and \<sigma>: "\<And> i x . funas_term (\<sigma> i x) \<subseteq> funas_trs R"
    and id: "shp = id"
  shows "\<not> SN (wfqrstep Q R)"
proof -
  from id have inj: "inj shp" by auto
  from chain have P_seq: "\<And>i. (s i, t i) \<in> DP shp R"
    and seq: "\<forall>i. ((t i) \<cdot> \<sigma> i, (s (Suc i)) \<cdot> \<sigma> (Suc i)) \<in> (qrstep Q R)^*"
    and NF: "\<And> i. s i \<cdot> \<sigma> i \<in> NF_terms Q"
    by (auto simp: ichain.simps)
  let ?s = "sharp_term shp"
  let ?us = "sharp_term (the_inv shp)"
  have "\<forall>i. \<exists>C. funas_ctxt C \<subseteq> funas_trs R \<and> (?us(s i), C\<langle>?us(t i)\<rangle>) \<in> R"
  proof
    fix i from P_seq have "(s i, t i) \<in> DP shp R" by simp
    from DP_step_in_R[OF this inj]  
    show "\<exists>C. funas_ctxt C \<subseteq> funas_trs R \<and> (?us (s i), C\<langle>?us(t i)\<rangle>) \<in> R" . 
  qed
  from choice[OF this] obtain C
    where C: "\<forall>i. funas_ctxt (C i) \<subseteq> funas_trs R \<and> (?us (s i), (C i)\<langle>?us(t i)\<rangle>) \<in> R" by best
  have wf_ctxt: "\<forall> i. funas_ctxt (C i) \<subseteq> funas_trs R" using C by auto
  {
    fix i
    have R: "(?us(s i), (C i)\<langle>?us(t i)\<rangle>) \<in> R" using C by auto
    have root_steps:
      "(?us(s i) \<cdot> (\<sigma> i), (C i \<cdot>c (\<sigma> i))\<langle>?us(t i) \<cdot> (\<sigma> i)\<rangle>) \<in> qrstep Q R"
      by (rule qrstepI[OF NF_imp_subt_NF R, of "(\<sigma> i)" _ _ Hole],
      insert NF[of i], auto simp: id F sharp_sig_def)
  } note root_steps = this
  let ?C = "D C \<sigma>"
  let ?s = "\<lambda>i. (?C i)\<langle>?us(s i) \<cdot> (\<sigma> i)\<rangle>"
  let ?t = "\<lambda>i. (?C (Suc i))\<langle>?us(t i) \<cdot> (\<sigma> i)\<rangle>"
  { 
    fix i
    from id have "sharp_term shp (s i) \<cdot> \<sigma> i = sharp_term shp (s i \<cdot> \<sigma> i)"
      by simp
  } note si = this
  { 
    fix i
    from id have "sharp_term shp (t i) \<cdot> \<sigma> i = sharp_term shp (t i \<cdot> \<sigma> i)"
      by simp
  } note ti = this
  {
    fix i
    have ctxt_steps: "(?s i, ?t i) \<in> qrstep Q R" using qrstep.ctxt[OF root_steps[of i], of "D C \<sigma> i"] unfolding D.simps ctxt_compose_apply_distrib .
  } note siti = this
  have aux: "\<forall>i.
    ( (?us(t i) \<cdot> \<sigma> i), (?us(s (Suc i)) \<cdot> \<sigma> (Suc i))) \<in> (qrstep Q R)^*"
    (is "\<forall>i. ?Q i") using seq si ti unfolding id the_inv_id by auto
  have "\<forall>i. (?us(t i) \<cdot> \<sigma> i, ?us(s (Suc i)) \<cdot> \<sigma> (Suc i)) \<in> (qrstep Q R)^*"
    (is "\<forall>i. ?P i")
  proof
    fix i
    from aux have "?Q i" by simp
    thus "?P i"
      unfolding F map_subst_distrib o_def  sharp_sig_def  id by simp 
  qed
  hence steps: "\<forall>i. (?t i, ?s (Suc i)) \<in> (qrstep Q R)^*" using 
    one_step_ctxt_closed_ctxt[OF rtrancl_one_step_ctxt_closed[OF ctxt_closed_imp_one_step_ctxt_closed[OF ctxt_closed_qrstep[of Q R]]]] by blast
  have big_step: "\<forall>i. (?s i, ?s(Suc i)) \<in> (qrstep Q R)^+"
  proof
    fix i
    from siti have "(?s i, ?t i) \<in> qrstep Q R" by best
    moreover from steps have "(?t i, ?s (Suc i)) \<in> (qrstep Q R)^*" by blast
    ultimately show "(?s i, ?s (Suc i)) \<in> (qrstep Q R)^+" by (rule rtrancl_into_trancl2)
  qed
  have "\<forall>i. funas_ctxt (?C i) \<subseteq> funas_trs R" using D_subst_wf[OF \<sigma> wf_ctxt]
    by auto
  moreover have "\<forall>i. funas_term (?us(s i) \<cdot> (\<sigma> i)) \<subseteq> funas_trs R"
  proof
    fix i
    from lhs_DP_wf and P_seq have "funas_term (s i) \<subseteq> F" unfolding F by best
    from this[unfolded F sharp_sig_def]
    have wf: "funas_term (?us(s i)) \<subseteq> funas_trs R" unfolding id by simp
    show "funas_term (?us(s i) \<cdot> \<sigma> i) \<subseteq> funas_trs R" unfolding funas_term_subst
      using wf vars_sharp_eq_vars \<sigma> by auto
  qed
  ultimately have "\<forall>i. funas_term (?s i) \<subseteq> funas_trs R" by auto
  with big_step
    have seq: "\<forall>i. (?s i, ?s(Suc i)) \<in> {(x, y) | x y. (x, y) \<in> (qrstep Q R)^+ \<and> funas_term x \<subseteq> funas_trs R \<and> funas_term y \<subseteq> funas_trs R}"
    (is "\<forall>i. _ \<in> ?steps") by blast
  have"?s 0 \<in> UNIV" by simp
  with seq have "?s 0 \<in> UNIV \<and> (\<forall>i. (?s i, ?s (Suc i)) \<in> ?steps)" by best
  hence "\<exists>S. S 0 \<in> UNIV \<and> (\<forall>i. (S i, S (Suc i)) \<in> ?steps)" by best
  hence nSN: "\<not> SN ?steps" unfolding SN_defs by blast
  show ?thesis
  proof (cases "wwf_qtrs Q R")
    case True
    from nSN have "\<not> SN ((wfqrstep Q R)^+)" unfolding wfqrstep_trancl[OF True] .
    with SN_trancl_SN_conv show ?thesis by auto
  next
    case False
    with SN_wfqrstep_imp_wwf_trs[of Q R]
    show ?thesis by blast
  qed
qed

hide_const D

fun funas_dpp :: "('f, 'v) dpp \<Rightarrow> 'f sig" where
  "funas_dpp (P, Pw, Q, Rw, R) = funas_trs P \<union> funas_trs Pw \<union> funas_trs R \<union> funas_trs Rw"

lemma ichain_imp_clean_ichain:
  fixes P R::"('f, 'v) trs" and F::"'f sig"
  defines "c \<equiv> \<lambda>t::('f, 'v) term. clean_term F t"
  defines "cs \<equiv> \<lambda>\<sigma>::('f, 'v) subst. clean_subst F \<sigma>"
  assumes subset: "funas_dpp (P, Pw, {}, {}, R) \<subseteq> F"
    and ichain: "ichain (P, Pw, {}, {}, R) s t \<sigma>"
  shows "ichain (P, Pw, {}, {}, R) s t (\<lambda>i. cs (\<sigma> i))"
proof -
  let ?P = "P \<union> Pw"
  from ichain have dpstep: "\<forall>i. (s i, t i) \<in> ?P"
    and seq: "\<forall>i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep R)^*"
    by (simp_all add: ichain.simps)
  from subset have subset_P: "funas_trs ?P \<subseteq> F" by (auto simp: funas_trs_def)
  from subset have subset_R: "funas_trs R \<subseteq> F" by simp
  {
    fix i
    from dpstep have "(s i, t i) \<in> ?P" by simp
    from lhs_wf[OF this subset_P] have "funas_term (s i) \<subseteq> F" .
  } note wf_s = this
  {
    fix i
    from dpstep have "(s i, t i) \<in> ?P" by simp
    from rhs_wf[OF this subset_P] have "funas_term (t i) \<subseteq> F" .
  } note wf_t = this
  have seq': "\<forall>i. (c (t i \<cdot> (\<sigma> i)), c (s (Suc i) \<cdot> (\<sigma> (Suc i)))) \<in> (rstep R)^*"
    (is "\<forall>i. (c (?t i), c (?s i)) \<in> _")
    using seq rsteps_imp_clean_rsteps[OF subset_R] unfolding c_def by blast
  have "\<forall>i. (t i \<cdot> cs (\<sigma> i), s (Suc i) \<cdot> cs (\<sigma> (Suc i))) \<in> (rstep R)^*"
  proof
    fix i
    have cts: "c (t i \<cdot> \<sigma> i) = t i \<cdot> cs (\<sigma> i)"
      using clean_subst_apply_term[OF wf_t[of i]]
      unfolding c_def cs_def o_def by auto
    have css: "c (s (Suc i) \<cdot> \<sigma> (Suc i)) = s (Suc i) \<cdot> cs (\<sigma> (Suc i))"
      using clean_subst_apply_term[OF wf_s[of "Suc i"]]
      unfolding c_def cs_def o_def by auto
    from seq'[THEN spec[of _ i]]
      show "(t i \<cdot> cs (\<sigma> i), s (Suc i) \<cdot> cs (\<sigma> (Suc i))) \<in> (rstep R)^*"
      unfolding cts css .
  qed
  with dpstep ichain show ?thesis by (simp add: ichain.simps)
qed

lemma left_linear_min_ichain_imp_clean_min_ichain:
  fixes F::"('f \<times> nat) set"
  defines "c \<equiv> clean_term F"
  defines "cs \<equiv> clean_subst F"
  assumes subset: "funas_dpp (P, Pw, {}, {}, R) \<subseteq> F"
    and left_linear: "left_linear_trs R"
    and chain: "min_ichain (P, Pw, {}, {}, R) s t \<sigma>"
  shows "min_ichain (P, Pw, {}, {}, R) s t (\<lambda>i. cs (\<sigma> i))"
proof -
  from chain have ichain: "ichain (P, Pw, {}, {}, R) s t \<sigma>"
    and minimal: "\<forall>i. SN_on (rstep R) {t i \<cdot> \<sigma> i}"
    by (auto simp: minimal_cond_def qrstep_rstep_conv)
  from ichain_imp_clean_ichain[OF subset ichain]
    have ichain: "ichain (P, Pw, {}, {}, R) s t (\<lambda> i. cs (\<sigma> i))" unfolding cs_def .
  from subset have "funas_trs R \<subseteq> F" by simp
  from subset have subset_P: "funas_trs (P \<union> Pw) \<subseteq> F" and subset_R: "funas_trs R \<subseteq> F" by (auto simp: funas_trs_def)
  {
    fix i
    from ichain have "(s i,t i) \<in> P \<union> Pw" by (simp add: ichain.simps)
    from rhs_wf[OF this subset_P] have "funas_term (t i) \<subseteq> F" .
  } note wf_t = this
  have minimal: "\<forall>i. SN_on (rstep R) {t i \<cdot> (cs(\<sigma> i))}"
  proof
    fix i show "SN_on (rstep R) {(t i) \<cdot> cs(\<sigma> i)}"
    proof -
      from 
        clean_subst_apply_term[of _ F, OF wf_t[of i], of "\<sigma> i"]
        left_linear_SN_on_imp_clean_SN_on[OF left_linear minimal[THEN spec[of _ i]], of F]
        show ?thesis unfolding cs_def c_def by simp
    qed
  qed
  with ichain show ?thesis by (auto simp: minimal_cond_def)
qed

lemma ichain_imp_clean_ichain_fcc: 
  fixes P Pw R::"('f, 'v) trs" and F::"'f sig"
  defines "c \<equiv> \<lambda>t::('f, 'v) term. clean_term F t"
  defines "cs \<equiv> \<lambda>\<sigma>::('f, 'v) subst. clean_subst F \<sigma>"
  assumes subset: "funas_args_trs (P \<union> Pw) \<union> funas_trs R \<subseteq> F"
  and non_var: "\<And>s t. (s,t) \<in> P \<union> Pw \<Longrightarrow> is_Fun s \<and> is_Fun t \<and> \<not> defined R (the (root t))"
  and nvar: "\<forall>(l, r)\<in>R. is_Fun l"
  and chain: "ichain (P,Pw,{},{},R) s t \<sigma>"
  shows "ichain (P,Pw,{},{},R) s t (\<lambda> i. cs (\<sigma> i))"
proof -
  from chain have dpstep: "\<And> i. (s i,t i) \<in> P \<union> Pw"
    and seq: "\<And> i. (t i \<cdot> \<sigma> i,s(Suc i)\<cdot>\<sigma>(Suc i)) \<in> (rstep R)^*"
    by (auto simp: ichain.simps)
  let ?split = "\<lambda> i f g ss ts. s i = Fun f ss \<and> t i = Fun g ts \<and> \<not> defined R (g, length ts)" 
  have fs: "\<forall> i. \<exists> f g ss ts. ?split i f g ss ts"
  proof
    fix i
    from non_var dpstep have "is_Fun (s i) \<and> is_Fun (t i)" by auto
    with non_var[OF dpstep[of i]] obtain f ss g ts where "?split i f g ss ts" by (cases "s i", auto, cases "t i", auto) 
    thus "\<exists> f g ss ts. ?split i f g ss ts" by blast
  qed
  from choice[OF fs] obtain f where fs: "\<forall> i. \<exists> g ss ts. ?split i (f i) g ss ts" ..
  from choice[OF fs] obtain g where fs: "\<forall> i. \<exists> ss ts. ?split i (f i) (g i) ss ts" ..
  from choice[OF fs] obtain ss where fs: "\<forall> i. \<exists> ts. ?split i (f i) (g i) (ss i) ts" ..
  from choice[OF fs] obtain ts where fs: "\<And> i. ?split i (f i) (g i) (ss i) (ts i)" by auto
  let ?s = "\<lambda> i. Fun (f i) (map c (ss i))"
  let ?t = "\<lambda> i. Fun (g i) (map c (ts i))"
  let ?ssj = "\<lambda> i j. ss i ! j \<cdot> \<sigma> i"
  let ?tsj = "\<lambda> i j. ts i ! j \<cdot> \<sigma> i"
  let ?ssc = "\<lambda> i j. ss i ! j \<cdot> cs(\<sigma> i)"
  let ?tsc = "\<lambda> i j. ts i ! j \<cdot> cs(\<sigma> i)"
  let ?cond = "\<lambda> i. length (ts i) = length (ss (Suc i)) \<and> g i = f (Suc i) \<and> (\<forall> j < length (ts i). (?tsj i j, ?ssj (Suc i) j) \<in> (rstep R)^*)"
  have argsteps: "\<And> i. ?cond i"
  proof -
    fix i
    from fs have ndef: "(g i, length (ts i)) \<notin> defs R" unfolding defs_def by auto
    from nondef_root_imp_arg_steps[OF seq[of i, simplified fs, simplified] nvar, simplified, OF ndef]
      nth_map[where xs = "ts  i"]
    show "?cond i" by auto
  qed
  from subset have subset_P: "funas_args_trs (P \<union> Pw) \<subseteq> F" by simp
  from subset have subset_R: "funas_trs R \<subseteq> F" by simp
  have main: "\<And> i j. j < length (ts i) \<Longrightarrow> (?tsc i j, ?ssc (Suc i) j) \<in> (rstep R)^*"
  proof -
    fix i j
    assume j: "j < length (ts i)"
    with fs have "?tsj i j \<in> set (args (t i \<cdot> \<sigma> i))" by simp
    from j fs[of "Suc i"] argsteps[of i] have "ss (Suc i) ! j \<in> set (args (s (Suc i)))" by auto 
    with dpstep[of "Suc i"] subset_P have wf_s: "funas_term (ss (Suc i) ! j) \<subseteq> F"
      unfolding funas_args_defs by force
    from j fs[of i] have "ts i ! j \<in> set (args (t i))" by auto 
    with dpstep[of i] subset_P have wf_t: "funas_term (ts i ! j) \<subseteq> F"
      unfolding funas_args_defs by force
    have c_t: "c(?tsj i j) = (?tsc i j)" unfolding c_def cs_def
      by (simp add: clean_subst_apply_term[OF wf_t])
    have c_s: "c(?ssj (Suc i) j) = (?ssc (Suc i) j)" unfolding c_def cs_def
      by (simp add: clean_subst_apply_term[OF wf_s])
    from argsteps[of i] j have "(?tsj i j, ?ssj (Suc i) j) \<in> (rstep R)^*" by simp
    from rsteps_imp_clean_rsteps[OF subset_R this] show "(?tsc i j, ?ssc (Suc i) j) \<in> (rstep R)^*"
      unfolding c_t[symmetric] c_s[symmetric] c_def .
  qed
  { 
    fix i
    from `?cond i` have len: "(length (ss (Suc i))) = length (ts i)" by simp
    have "(Fun (g i) (map (\<lambda> t. t \<cdot> cs (\<sigma> i)) (ts i)), Fun (g i) (map (\<lambda> t. t \<cdot> cs (\<sigma> (Suc i))) (ss (Suc i)))) \<in> (rstep R)^*" 
    proof (rule mp[OF spec[OF spec[OF spec[OF rstep_rtrancl_all_ctxt_closed[of R, unfolded all_ctxt_closed_def]]]] conjI],
      simp add: `?cond i`, simp add: len, intro allI impI)
      fix j
      assume "j < length (ts i)"
      from main[OF this]
      show "(?tsc i j, ?ssc (Suc i) j) \<in> (rstep R)^*" .
    qed
    with `?cond i` 
    have "(t i \<cdot> cs (\<sigma> i), s (Suc i) \<cdot> cs (\<sigma> (Suc i))) \<in> (rstep R)^*" by (simp add: fs[of i] fs[of "Suc i"])
  } 
  with dpstep chain show ?thesis by (auto simp: ichain.simps)
qed

(* it follows the crucial lemma for flat context closures *)
(* unfortunately, it does not hold for non-left-linear systems, see 
   CSL'10 paper *)
(* and even more unfortunately, it does not hold for strict R rules:
   P = {}, Pw = {F(x,y) \<rightarrow> F(x,f(x))}, Q = {}, R = {a \<rightarrow> b}, Rw = {f(a) \<rightarrow> f(a)}
   inf min-chain: F(g(a),f(g(b))) \<rightarrow>Pw F(g(a),f(g(a))) \<rightarrow>R F(g(a),f(g(b))) \<rightarrow>Pw ...
   however, if we restrict substitution to non-top signature of DPP which 
   here is {f,a,b}, then there is no inf. minimal chain:
   - if inf-chain, then there must be inf. many R-rules as P is empty.
   - if we can apply any R-rule, then \<sigma> x must contain a as subterm
   - with signature {f,a,b} \<Rightarrow> \<sigma> x = f^i(a)
   - f(x)\<sigma> contains f(a) as subterm, and hence f(x)\<sigma> is non-terminating w.r.t. Rw
   - hence, chain is not minimal
*)
lemma left_linear_min_ichain_imp_clean_min_ichain_fcc: 
  fixes P Pw R::"('f,'v)trs" and F::"'f sig"
  defines "c \<equiv> \<lambda>t::('f, 'v) term. clean_term F t"
  defines "cs \<equiv> \<lambda>\<sigma>::('f, 'v) subst. clean_subst F \<sigma>"
  assumes left_linear: "left_linear_trs R"
  and subset: "funas_args_trs (P \<union> Pw) \<union> funas_trs R \<subseteq> F"
  and non_var: "\<And>s t. (s,t) \<in> P \<union> Pw \<Longrightarrow> is_Fun s \<and> is_Fun t \<and> \<not> defined R (the (root t))"
  and chain: "min_ichain (P,Pw,{},{},R) s t \<sigma>"
  shows "min_ichain (P,Pw,{},{},R) s t (\<lambda> i. cs (\<sigma> i))"
proof -
  from chain have ichain: "ichain (P,Pw,{},{},R) s t \<sigma>"
    and min: "\<And> i. SN_on (rstep R) {t i \<cdot> \<sigma> i}" by (auto simp: minimal_cond_def)
  from left_var_imp_not_SN[where R = R and t = "t 0 \<cdot> \<sigma> 0"] min[of 0]
  have nvar: "\<forall>(l, r)\<in>R. is_Fun l" by force
  from ichain_imp_clean_ichain_fcc[OF subset non_var nvar ichain]
  have ichain: "ichain (P,Pw,{},{},R) s t (\<lambda> i. cs (\<sigma> i))" unfolding cs_def .
  from subset have subset_P: "funas_args_trs (P \<union> Pw) \<subseteq> F" unfolding funas_dpp.simps by auto
  from subset have subset_R: "funas_trs R \<subseteq> F" by simp
  {
    fix i
    from ichain have "(s i, t i) \<in> P \<union> Pw" by (simp add: ichain.simps)
    with subset_P non_var[OF this] obtain f ts where ti: "t i = Fun f ts" and tiP: "funas_args_term (Fun f ts) \<subseteq> F" and ndef: "(f, length (map (\<lambda> t. t \<cdot> cs (\<sigma> i)) ts)) \<notin> defs R" 
      unfolding funas_args_trs_def funas_args_rule_def by (cases "t i", simp, force simp: defs_def)
    have "SN_on (rstep R) {t i  \<cdot> (cs(\<sigma> i))}"
    proof (simp add: ti, rule Trs.SN_args_imp_SN[OF _ nvar ndef]) 
      fix u
      assume "u \<in> set (map (\<lambda> t. t \<cdot> cs (\<sigma> i)) ts)" (is "_ \<in> set ?ts")
      then obtain uu where uu: "uu \<in> set ts" and u: "u = uu \<cdot> cs (\<sigma> i)" by auto      
      from uu tiP have wuu: "funas_term uu \<subseteq> F" unfolding funas_args_term_def by auto
      from SN_imp_SN_arg[OF min[of i, simplified ti, simplified]] uu have "SN_on (rstep R) {uu \<cdot> \<sigma> i}" by auto
      from left_linear_SN_on_imp_clean_SN_on[OF left_linear this, of F] show "SN_on (rstep R) {u}"
        unfolding clean_subst_apply_term[OF wuu] u cs_def .
    qed
  }
  with ichain show ?thesis by (simp add: minimal_cond_def)
qed



lemma funas_dpp_initial_subset_sharp_sig:
  "funas_dpp (DP shp R, {}, Q,{}, R) \<subseteq> sharp_sig shp (funas_trs R)"
using funas_DP_subset unfolding sharp_sig_def by simp

lemma SN_wfrstep_imp_SN_rstep:
  assumes "SN (wfrstep R)" shows "SN (rstep R)"
proof (cases "wf_trs R")
  case False
  with SN_wfrstep_imp_wf_trs[OF assms] show ?thesis by blast
next
  case True
  show ?thesis
  proof (rule ccontr)
    let ?F = "sharp_sig id (funas_trs R)"
    let ?c = "clean_term ?F"
    let ?cs = "clean_subst ?F"
    assume "\<not> SN (rstep R)"
    from not_SN_imp_ichain_rstep[OF True this]
      obtain s t \<sigma> where ichain: "ichain (DP id R, {},{},{},R) s t \<sigma>" by blast
    from ichain_imp_clean_ichain[OF funas_dpp_initial_subset_sharp_sig this]
    have ichain: "ichain (DP id R,{},{},{},R) s t (\<lambda>i. ?cs(\<sigma> i))" .
    have  "\<not> SN (wfrstep R)" unfolding wfrstep_def qrstep_rstep_conv[symmetric]
    proof (rule clean_ichain_imp_not_SN_wfrstep[OF ichain _ refl, unfolded wfqrstep_def])
      fix i x
      show "funas_term (?cs (\<sigma> i) x) \<subseteq> funas_trs R" unfolding clean_subst.simps
        o_def using funas_term_clean_term[of "funas_trs R"] 
        unfolding sharp_sig_def by auto
    qed
    with assms show False by blast
  qed
qed

lemma SN_wfrstep_SN_rstep_conv: "SN (wfrstep R) = SN (rstep R)"
proof -
  {
    assume "\<not> SN (wfrstep R)"
    then obtain S where "\<forall>i. (S i, S (Suc i)) \<in> wfrstep R" unfolding SN_defs by auto
    hence "\<not> SN (rstep R)" unfolding SN_defs wfrstep_def by blast
  }
  with SN_wfrstep_imp_SN_rstep show ?thesis by blast
qed


lemma SN_wf_ground: assumes c: "(c,0) \<in> F" and F: "funas_trs R \<subseteq> F"
  and SN: "\<And> t. funas_term t \<subseteq> F \<Longrightarrow> ground t \<Longrightarrow> SN_on (rstep R) {t}"
  shows "SN (rstep R)"
proof (rule SN_wfrstep_imp_SN_rstep, rule)
  fix f 
  assume steps: "\<forall> i. (f i, f (Suc i)) \<in> wfrstep R"
  hence f0: "funas_term (f 0) \<subseteq> funas_trs R" and steps: "\<And> i. (f i, f (Suc i)) \<in> rstep R" unfolding wfrstep_def by auto
  let ?\<sigma> = "(\<lambda> _. Fun c [])"
  from f0 c have wf: "funas_term (f 0 \<cdot> ?\<sigma>) \<subseteq> F" (is "funas_term ?t \<subseteq> _") unfolding funas_term_subst using F by auto
  have "ground ?t" by auto
  from SN[OF wf this] have "SN_on (rstep R) {?t}" by auto
  from SNinstance_imp_SN[OF this] steps
  show False unfolding SN_defs by auto
qed

end